#include <iostream>
#include <typeinfo>
using namespace std;

class Basis
{
    public:
    virtual void gebeAus(void)
    {
        cout << "Basis" << endl;
    }
};

class Ableitung :
public Basis
{
    public:
    virtual void gebeAus(void) // wre auch "virtual", wenn man das nicht dazuschriebe
    {
        cout << "Ableitung" << endl;
    }
    void tuWasSpezielles()
    {
        cout << "was Spezielles" << endl;
    }
};

int main (int argc, char* (arv[]) )
{
    // dynamische Typumwandlung ******************************

    Basis* b = new Ableitung (); // implizite Aufwrtskonvertierung: "Upcast"
    b->gebeAus(); // Gibt den Text "Ableitung" aus, dank Polymorphie.

    // b->tuWasSpezielles(); // Das ginge im Prinzip fr _dieses_ *b,
                             // ist aber verboten, weil die Klasse Basis das nicht kann.

    // Idee:     Ableitung* c = (Ableitung*) (b);
    // Aber da macht C++ anders als Java keine Typprfung.
    // So ists besser:

    Ableitung* c = dynamic_cast<Ableitung*> (b); // Abwrtskonvertierung: "Downcast"

    if(c != 0) // dynamic_cast liefert Nullzeiger, falls keine Konvertierung mglich
    {
        c->gebeAus();
    }
    else
    {
        cout << "kein dynamic_cast von Zeiger moeglich" << endl;
    }



    // dasselbe mit Referenzen ******************************

    Ableitung d;
    Basis& e = d; // implizite Aufwrtskonvertierung: "Upcast"
    e.gebeAus(); // Gibt den Text "Ableitung" aus, dank Polymorphie.

    // e.tuWasSpezielles(); // Das ginge im Prinzip fr _dieses_ e, ist aber verboten,
                            // weil die Klasse Basis keine solche Methode hat.

    try
    {
        Ableitung& f = dynamic_cast<Ableitung&> (e); // Abwrtskonvertierung: "Downcast"
        f.gebeAus();
    }
    catch(bad_cast ex) // Fehlgeschlagene Umwandlung von Referenzen wirft Exception.
    {
        cout << "kein dynamic_cast von Referenz moeglich" << endl;
    }


    // Anmerkung: Damit dynamic_cast funktioniert, brauchen die Klassen mindestens eine virtuelle Methode.
    // Das ist aber blicherweise kein Problem, weil man sowieso schon den Destruktor virtuell machen wird.

    // Bei Microsoft Visual C++ 6.0 muss man auerdem mit der Compiler-Option /GR
    // die Run-Time Type Identification (RTTI) anstellen.
    // dynamic_cast mit Referenzen wirft dort anders als dokumentiert zwar eine Exception,
    // aber keine vom Typ bad_cast.


    // andere "Casts" ******************************

    // Gewhnliche Konvertierungen
    double g = 3.14;
    int h;
    h = g; // implizite Konvertierung mit Gefahr kaputter Daten, geht in Java nicht
    h = (int)g; // genauso gefhrlich, aber man siehts wenigstens, wie in Java
    h = static_cast<int> (g); // Schreibweise in Standard-C++, immer noch gefhrlich;
                              // "statisch", weil keine Typberprfung zur Laufzeit.
                              // Beachte: Die Bits von g werden ggf. umgewandelt!

    char* i = reinterpret_cast<char*> (&g); // Fasse die gleichen Bits anders auf.
                                            // _Kann_ bei Ein- und Ausgabe helfen.
                                            // Ist aber _sehr_ gefhrlich.




    cin.get(); // warte auf Return-Taste
    return 0;
}
